/*
     This file is part of CPU
     (C) 2003 Blake Matheny (and other contributing authors)

     CPU is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     CPU is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with CPU; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * Initialization routines and some helper routines
 * @author Blake Matheny
 * @file ld.c
 **/
#include <stdio.h>
#include <ldap.h>
#include <string.h>
#include "plugins/ldap/ldap.h"

static int list_size = 0;

const char * ldap_hashes[] = {
  "{sha}",
  "{ssha}",
  "{md5}",
  "{smd5}",
  "{crypt}",
  NULL,
};

int
ldapOperation(ldapop_t optype)
{
  LDAP * ld;

  ld = ldap_open((char*)globalLdap->hostname, (int)globalLdap->port);
  if ( ld == NULL )
    {
      ldap_perror(ld, "ldap: ldapOperation: ldap_open");
      return -1;
    }

  if ( ldap_bind_s(ld, globalLdap->bind_dn, globalLdap->bind_password,
	           LDAP_AUTH_SIMPLE) != LDAP_SUCCESS )
    {
      ldap_perror(ld, "ldap: ldapOperation: ldap_bind_s");
      return -1;
    }

  switch(optype)
    {
      case USERADD:
	if ( (int)globalLdap->passent->pw_uid < 0 )
	  {
	    if ( (int)(globalLdap->passent->pw_uid = getNextUid(ld)) < 0)
	      return -1;
	  }
	if ( (int)globalLdap->passent->pw_gid < 0 )
	  {
	    if ( (int)(globalLdap->passent->pw_gid = getNextGid(ld, USERADD))
	         < 0)
	      return -1;
	  }
	if (ldapUserAdd(ld) < 0)
	  return -1;
	break;
      case USERMOD:
	if (ldapUserMod(ld) < 0)
	  return -1;
	break;
      case USERDEL:
	if (ldapUserDel(ld) < 0)
	  return -1;
	break;
      case GROUPADD:
	if ( (int)globalLdap->passent->pw_gid < 0 )
	  {
	    if ( (int)(globalLdap->passent->pw_gid = getNextGid(ld, GROUPADD))
	         < 0)
	      return -1;
	  }
	if (ldapGroupAdd(ld) < 0)
	  return -1;
	break;
      case GROUPMOD:
	if (ldapGroupMod(ld) < 0)
	  return -1;
	break;
      case GROUPDEL:
	if (ldapGroupDel(ld) < 0)
	  return -1;
	break;
      case CAT:
	if (ldapCat(ld) < 0)
	  return -1;
	break;
      default:
	fprintf(stderr, "ldap: ldapOperation: Unknown optype\n");
	return -1;
	break;
    }
  return 0;
}

#define ALLOC_SIZE 1024

LDAPMod **
ldapAddList(LDAPMod ** mods)
{
  LDAPMod ** tmod = NULL;
  int lsize = 0;
  int i = 0;
  lsize = list_size+2;

  tmod = (LDAPMod**)malloc(sizeof(LDAPMod*)*(lsize));
  if ( tmod == NULL )
    return NULL;
  bzero(tmod, sizeof(LDAPMod*)*(lsize));

  if ( mods != NULL )
    {
      for ( i = 0; mods[i] != NULL; i++ )
	tmod[i] = mods[i];
      //free(mods);
    }

  tmod[list_size] = (LDAPMod*)malloc(sizeof(LDAPMod));
  bzero(tmod[list_size], sizeof(LDAPMod));

  tmod[list_size+1] = NULL;

  /*
  mods = (LDAPMod**)realloc(mods, (list_size+ALLOC_SIZE)*sizeof(LDAPMod*));
  mods[list_size] = (LDAPMod*)malloc(sizeof(LDAPMod));
  bzero(mods[list_size], sizeof(LDAPMod));
  */
  return tmod;
}

LDAPMod **
ldapBuildListStr(int mod_op, char * mod_type, char * value, LDAPMod ** mods)
{
  char **temp;

  if ( value == NULL )
    return mods;

  mods = ldapAddList(mods);

  temp = (char**)malloc(sizeof(char*)*2);
  bzero(temp, sizeof(char*)*2);
  temp[0] = value;
  temp[1] = NULL;

  mods[list_size]->mod_op = mod_op;
  mods[list_size]->mod_type = strdup(mod_type);
  mods[list_size]->mod_values = temp;
  list_size++;
  return mods;
}

LDAPMod **
ldapBuildList(int mod_op, char * mod_type, char ** value, LDAPMod ** mods)
{
  mods = ldapAddList(mods);

  if ( value == NULL )
    return mods;

  mods[list_size]->mod_op = mod_op;
  mods[list_size]->mod_type = strdup(mod_type);
  mods[list_size]->mod_values = value;
  list_size++;
  return mods;
}

LDAPMod **
ldapBuildListInt(int mod_op, char * mod_type, int value, LDAPMod ** mods)
{
  char **temp;
  size_t len = 1;
  int valtemp = ABS(value);

  mods = ldapAddList(mods);
  temp = (char**)malloc(sizeof(char*)*2);
  bzero(temp, sizeof(char*)*2);

  while ( valtemp/10 > 0 )
    {
      len++;
      valtemp = valtemp/10;
    }

  if ( value < 0 )
    len++;
  len++;
  temp[0] = (char*)malloc(sizeof(char)*len);
  bzero(temp[0], sizeof(char)*len);
  snprintf(temp[0], len, "%d", value);
  temp[1] = NULL;

  mods[list_size]->mod_op = mod_op;
  mods[list_size]->mod_type = strdup(mod_type);
  mods[list_size]->mod_values = temp;
  list_size++;
  return mods;
}

char *
ldapGetCn(void)
{
  size_t slen = 0;
  char * temp;

  if (globalLdap->first_name && globalLdap->last_name)
    {
      slen =
	strlen (globalLdap->first_name) +
	strlen (globalLdap->last_name) + 2;
      temp = (char *) malloc (sizeof (char) * slen);
      if (temp == NULL)
	return NULL;
      bzero (temp, slen * sizeof (char));
      snprintf (temp, slen, "%s %s", globalLdap->first_name,
		globalLdap->last_name);
    }
  else if (globalLdap->first_name)
    temp = globalLdap->first_name;
  else if (globalLdap->last_name)
    temp = globalLdap->last_name;
  else
    temp = globalLdap->passent->pw_name;

  return temp;
}

const char *
ldapGetHashPrefix(hash_t hasht)
{
  switch(hasht)
    {
      case H_SHA1:
	return ldap_hashes[H_SHA1];
	break;
      case H_SSHA1:
	return ldap_hashes[H_SSHA1];
	break;
      case H_MD5:
	return ldap_hashes[H_MD5];
	break;
      case H_SMD5:
	return ldap_hashes[H_SMD5];
	break;
      case H_CRYPT:
	return ldap_hashes[H_CRYPT];
	break;
      case H_CLEAR:
	/* FIXME: this should work so that the prefix is returned for the
	   correct hash but the password doesn't get encrypted */
	return ldap_hashes[H_CRYPT];
	break;
      default:
	fprintf(stderr, "ldap: ldapGetHashPrefix: Unknown hash type.\n");
	break;
    }
  return NULL;
}

uid_t
getNextUid(LDAP * ld)
{
  LDAPMessage * res[2];
  uid_t min_uid;
  uid_t next_uid = 0;
  uid_t max_uid;
  int i = 0;
  int id_pass = 0;
  int random = 0;
  int ldapres = 0;
  int filtsize = 13;
  int digitsize[2];
  struct timeval timeout;
  char * filter = NULL;
  char * temp;
  char * attrs[2] = {
    "uidNumber",
    NULL
  };

  timeout = globalLdap->timeout;
  res[1] = NULL;
  digitsize[0] = digitsize[1] = 0;

  temp = cfg_get_str("LDAP", "RANDOM");
  if ( temp == NULL )
    random = 0;
  else if ( tolower(temp[0]) == 'f' )
    random = 0;
  else
    random = 1;

  id_pass = cfg_get_int("LDAP", "ID_MAX_PASSES");
  if ( id_pass < 1 )
    id_pass = 1000;

  min_uid = cfg_get_int("LDAP", "MIN_UIDNUMBER");
  if ( min_uid < 0 )
    min_uid = 100;

  max_uid = cfg_get_int("LDAP", "MAX_UIDNUMBER");
  if ( max_uid > 1000000 )
    max_uid = 10000;

  digitsize[0] = ((min_uid/10)+1);
  while(i < id_pass)
    {
      if ( !random )
	next_uid = min_uid+i;
      else
	next_uid = cRandom(min_uid, max_uid);
      digitsize[0] = (next_uid/10)+1;
      if ( digitsize[0] > digitsize[1] )
	{
	  filtsize += digitsize[0];
	  Free(filter);
	  filter = (char*)malloc(sizeof(char)*filtsize);
	  if ( filter == NULL )
	    return -1;
	  bzero(filter, filtsize);
	  snprintf(filter, filtsize, "(uidNumber=%d)", next_uid);
	  digitsize[1] = digitsize[0];
	}
      else
	{
	  bzero(filter, filtsize);
	  snprintf(filter, filtsize, "(uidNumber=%d)", next_uid);
	}
      if ( ldap_search_st(ld, globalLdap->user_base, LDAP_SCOPE_SUBTREE,
			  filter, attrs, 0, &timeout, res)
	  != LDAP_SUCCESS )
	{
	  ldap_perror(ld, "ldap: getNextUid: ldap_search_st");
	  return -1;
	}
      ldapres = ldap_count_entries(ld, res[0]);
      if ( ldapres == 0 )
	break;
      i++;
    }
  if ( i == id_pass )
    {
      fprintf(stderr, "ldap: getNextUid: Unable to find new uid.\n");
      return -1;
    }
  return next_uid;
}

gid_t
getNextGid(LDAP * ld, ldapop_t op)
{
  LDAPMessage * res[2];
  gid_t min_gid;
  gid_t next_gid = 0;
  gid_t max_gid;
  int i = 0;
  int id_pass = 0;
  int random = 0;
  int ldapres = 0;
  int filtsize = 13;
  int digitsize[2];
  struct timeval timeout;
  char * filter = NULL;
  char * temp;
  char * attrs[7] = {
    "gidNumber",
    NULL
  };

  timeout = globalLdap->timeout;
  res[1] = NULL;
  digitsize[0] = digitsize[1] = 0;

  temp = cfg_get_str("LDAP", "RANDOM");
  if ( temp == NULL )
    random = 0;
  else if ( tolower(temp[0]) == 'f' )
    random = 0;
  else
    random = 1;

  id_pass = cfg_get_int("LDAP", "ID_MAX_PASSES");
  if ( id_pass < 1 )
    id_pass = 1000;

  min_gid = cfg_get_int("LDAP", "MIN_GIDNUMBER");
  if ( min_gid < 0 )
    min_gid = 100;

  max_gid = cfg_get_int("LDAP", "MAX_GIDNUMBER");
  if ( max_gid > 1000000 )
    max_gid = 10000;

  digitsize[0] = ((min_gid/10)+1);
  while(i < id_pass)
    {
      if ( !random )
	next_gid = min_gid+i;
      else
	next_gid = cRandom(min_gid, max_gid);
      digitsize[0] = (next_gid/10)+1;
      if ( digitsize[0] > digitsize[1] )
	{
	  filtsize += digitsize[0];
	  Free(filter);
	  filter = (char*)malloc(sizeof(char)*filtsize);
	  if ( filter == NULL )
	    return -1;
	  bzero(filter, filtsize);
	  snprintf(filter, filtsize, "(gidNumber=%d)", next_gid);
	  digitsize[1] = digitsize[0];
	}
      else
	{
	  bzero(filter, filtsize);
	  snprintf(filter, filtsize, "(gidNumber=%d)", next_gid);
	}
      if ( op == USERADD )
	{
	  if ( ldap_search_st(ld, globalLdap->user_base, LDAP_SCOPE_SUBTREE,
		              filter, attrs, 0, &timeout, res)
	       != LDAP_SUCCESS )
	    {
	      ldap_perror(ld, "ldap: getNextGid: ldap_search_st");
	      return -1;
	    }
	}
      else if ( op == GROUPADD )
	{
	  if ( ldap_search_st(ld, globalLdap->group_base, LDAP_SCOPE_SUBTREE,
	                      filter, attrs, 0, &timeout, res)
	       != LDAP_SUCCESS )
	    {
	      ldap_perror(ld, "ldap: getNextGid: ldap_search_st");
	      return -1;
	    }
	}
      else
	return -1;
      ldapres = ldap_count_entries(ld, res[0]);
      if ( ldapres == 0 )
	break;
      i++;
    }
  if ( i == id_pass )
    {
      fprintf(stderr, "ldap: getNextGid: Unable to find new gid.\n");
      return -1;
    }
  return next_gid;
}
